from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.parser import VideoParser
from moviepy.video.fx import MirrorY
from moviepy import Clip


class FlipVerticallyEffect(Effect):
    """
    This effect flips the video vertically (including
    the mask).
    """
    def apply(self, video: Clip) -> Clip:
        # TODO: This is not working properly yet
        #PythonValidator.validate_method_params(BlinkEffect.apply, locals(), ['video'])
        video = VideoParser.to_moviepy(video)

        return MirrorY().apply(video)
    
    # TODO: I don't need this
    def apply_over_video(self, video, background_video):
        return super().apply_over_video(video, background_video)


from yta_multimedia.video.edition.settings import SMALL_AMOUNT_TO_FIX


class VideoFrameTHelper:
    """
    Class to simplify and encapsulate the conversion
    between video and audio time frame moments 't' and
    frame indexes.
    """

    @staticmethod
    def get_number_of_frames(duration: float, fps: float):
        """
        Get the numbers of frames with the given 'duration'
        and 'fps'.
        """
        return int(duration * fps)

    @staticmethod
    def get_video_frames_indexes_from_duration_and_fps(duration: float, fps: float):
        """
        Get all the video frame indexes for a video with
        the given 'fps' and 'duration'. 
        """
        return [
            i
            for i in range(fps * duration)
        ]
    
    @staticmethod
    def get_video_frames_indexes_from_number_of_frames(number_of_frames: int):
        """
        Get all the video frame indexes for a video with
        the given 'number_of_frames'.
        """
        return [
            i
            for i in range(number_of_frames)
        ]

    @staticmethod
    def get_video_frames_ts_from_duration_and_fps(duration: float, fps: float):
        """
        Get all the 't' frames time moments for a video
        with the given 'fps' and 'duration'. Each 't'
        includes a small amount increased to ensure it
        fits the frame time range.
        """
        return [
            VideoFrameTHelper.get_frame_t_from_frame_index(i, fps)
            for i in VideoFrameTHelper.get_video_frames_indexes_from_duration_and_fps(duration, fps)
        ]
    
    @staticmethod
    def get_video_frames_ts_from_number_of_frames(number_of_frames: int, fps: float):
        """
        Get all the 't' frames time moments for a video
        with the given 'number_of_frames'. Each 't'
        includes a small amount increased to ensure it
        fits the frame time range.
        """
        return [
            VideoFrameTHelper.get_frame_t_from_frame_index(i, fps)
            for i in VideoFrameTHelper.get_video_frames_indexes_from_number_of_frames(number_of_frames)
        ]
    # TODO: Should I move this above to another class (?)

    @staticmethod
    def get_frame_t_base(t: float, fps: float):
        """
        Turn the provided 't' video frame time moment to
        the real base one (the one who is the start of
        the frame time interval, plus a minimum quantity
        to avoid floating point number issues).
        """
        return VideoFrameTHelper.get_frame_index_from_frame_t(t, fps) / fps + SMALL_AMOUNT_TO_FIX
    
    @staticmethod
    def get_frame_index_from_frame_t(t: float, fps: float):
        """
        Get the video audio frame 't' time moment base,
        which is the left (lower) limit of that 't' time
        moment.

        As a reminder, any 't' time moment goes from the
        lower limit (including it) to the upper limit 
        (not included). So, there is a time interval in 
        which any 't' time moment is included, which is
        defined by the range [lower_limit, upper_limit).

        For example, in a video with fps=30, any value
        between [0/30, 1/30) will be recognized as a 't'
        time moment for the frame index 0. As you can see,
        the 1/30 is not included, because it will be part
        of the next index, as it is its lower limit that
        is included.
        """
        frame_duration = 1 / fps

        return int((t + SMALL_AMOUNT_TO_FIX) // frame_duration)

    @staticmethod
    def get_frame_t_from_frame_index(index: int, fps: float):
        """
        Get the frame time moment t from the given
        frame 'index'.
        """
        return index / fps + SMALL_AMOUNT_TO_FIX
    
    @staticmethod
    def get_video_audio_tts_from_video_frame_t(video_t: float, video_fps: float, audio_fps: float):
        """
        Get all the audio time moments associated to
        the given 'video' 't' time moment, as an array.

        One video time moment 't' is associated with a lot
        of video audio time 't' time moments. The amount 
        of video audio frames per video frame is calculated
        with the divions of the audio fps by the video fps.

        The result is an array of 't' video audio time
        moments. Maybe you need to turn it into a numpy
        array before using it as audio 't' time moments.
        """
        from yta_general_utils.math.progression import Progression

        audio_frames_per_video_frame = int(audio_fps / video_fps)
        audio_frame_duration = 1 / audio_fps
        video_frame_duration = 1 / video_fps

        t = VideoFrameTHelper.get_frame_t_base(video_t, video_fps)

        return Progression(t, t + video_frame_duration - audio_frame_duration, audio_frames_per_video_frame).values

    @staticmethod
    def get_video_frame_t_from_video_audio_frame_t(audio_t: float, video_fps: float):
        """
        Get the video frame time moment t from the given
        video audio frame time moment 'audio_t'.
        """
        return VideoFrameTHelper.get_frame_t_base(audio_t , video_fps)

    @staticmethod
    def get_video_frame_index_from_video_audio_frame_index(audio_index: int, video_fps: float, audio_fps: float):
        """
        Get the video frame index from the given video
        audio frame index 'audio_index'.
        """
        return round(audio_index * (video_fps / audio_fps))

    @staticmethod
    def get_video_frame_t_from_video_audio_frame_index(audio_index: int, video_fps: float, audio_fps: float):
        """
        Get the video frame time moment t from the given
        video audio frame index 'audio_index'.
        """
        return VideoFrameTHelper.get_frame_t_from_frame_index(
            VideoFrameTHelper.get_video_frame_index_from_video_audio_frame_index(
                audio_index,
                video_fps,
                audio_fps
            ),
            video_fps
        )
    
    @staticmethod
    def get_video_frame_index_from_video_audio_frame_t(audio_t: float, video_fps: float, audio_fps: float):
        """
        Get the video frame index from the given video
        audio frame time moment 'audio_t'.
        """
        return VideoFrameTHelper.get_video_frame_index_from_video_audio_frame_index(
            VideoFrameTHelper.get_frame_index_from_frame_t(
                audio_t,
                audio_fps
            ),
            video_fps,
            audio_fps
        )


from yta_general_utils.math.rate_functions.rate_function_argument import RateFunctionArgument
from random import randint
from abc import abstractmethod
from math import pi

import numpy as np


class PositionAdditionalMovement: 
    """
    The additional movement that can be added to
    an element in a position. This will move the
    element in some way but in the position in
    what it is.
    """  
    @abstractmethod
    def get_position(self):
        """
        Obtain the position on the given 'n' moment of
        the movement ('n' value must be a normalized
        value between 0 and 1 representing the amount
        of movement done for the moment to be
        calculated where 0 is the start and 1 the end).
        """
        pass

class ShakeAdditionalMovement(PositionAdditionalMovement):
    """
    Make the element shake at the same speed in
    the position it is.
    """
    def __init__(self, shake_speed: int = 4):
        self.shake_speed = 4 if shake_speed is None else shake_speed

    def get_position(
        self,
        position: tuple[int, int],
    ):
        return shake(position, self.shake_speed)
    
class ShakeIncreasingAdditionalMovement(PositionAdditionalMovement):
    """
    Make the element shake in an increasing speed
    in the position it is.
    """
    def __init__(self, shake_speed: int = 4):
        self.shake_speed = 4 if shake_speed is None else shake_speed

    def get_position(
        self,
        n: float,
        position: tuple[int, int],
        rate_function: RateFunctionArgument = RateFunctionArgument.default()
    ):
        return shake_increasing(position, self.shake_speed, n, rate_function)
    
class ShakeDecreasingAdditionalMovement(PositionAdditionalMovement):
    """
    Make the element shake in a decreasing speed
    in the position it is.
    """
    def __init__(self, shake_speed: int = 4):
        self.shake_speed = 4 if shake_speed is None else shake_speed

    def get_position(
        self,
        n: float,
        position: tuple[int, int],
        rate_function: RateFunctionArgument = RateFunctionArgument.default()
    ):
        return shake_decreasing(position, self.shake_speed, n, rate_function)
        
class CirclesAdditionalmovement(PositionAdditionalMovement):
    """
    Make the element move in circles around the
    position in which it is.
    """
    def __init__(self, time_per_circle: float = 1, radius: int = 200):
        self.time_per_circle = 1 if time_per_circle is None else time_per_circle
        self.radius = 200 if radius is None else radius

    def get_position(
        self,
        n: float,
        position: tuple[int, int],
        rate_function: RateFunctionArgument = RateFunctionArgument.default()
    ):
        return circles(position, self.time_per_circle, self.radius, n, rate_function)

# TODO: Remove this in the next commit
# class AdditionalMovement(Enum):
#     """
#     The additional movement that can be added to
#     an element in a position. This will move the
#     element in some way but in the position in
#     what it is.
#     """

#     SHAKE = 'shake'
#     """
#     Make the element shake at the same speed in
#     the position it is.
#     """
#     SHAKE_INCREASING = 'shake_increasing'
#     """
#     Make the element shake in an increasing speed
#     in the position it is.
#     """
#     SHAKE_DECREASING = 'shake_decreasing'
#     """
#     Make the element shake in a decreasing speed
#     in the position it is.
#     """

#     def get_function(self):
#         """
#         Obtain the function to calculate the value of
#         the given 'n'. This function can be called by
#         providing the 'n' value and any other needed
#         key parameter.
#         """
#         return {
#             self.SHAKE: shake,
#             self.SHAKE_INCREASING: shake_increasing_movement,
#             self.SHAKE_DECREASING: shake_decreasing_movement
#         }[self]

#     def apply(self, n: float, **kwargs) -> float:
#         """
#         Apply the additional movement to the given 'n'
#         and kwargs.

#         The 'n' parameter must be a normalized value
#         (a value between 0 and 1).

#         The result will be the corresponding 'y' 
#         value for the given 'n' value.
#         """
#         if not NumberValidator.is_number_between(n, 0, 1):
#             raise Exception('The additional movement has been built to work with an "n" value between 0 and 1.')

#         return self.get_function()(n, **kwargs)

"""
        In place movements (static) effect position functions below
"""
def shake(
    position: tuple[int, int],
    shake_speed: int
):
    """
    Creates a shake effect on the given 'position',
    by moving the element in a random position,
    that is constant during the time.

    The 'shake_speed' is the amount of pixels to
    move in each movement. The value 4 is a good
    value.

    The 'n' parameter must be a normalized value
    between 0 and 1 representing the amount of
    animation that has been processed until the
    moment in which this method is executed.
    """
    # TODO: I think this is increasing also not a
    # constant shaking movement because of the *
    #shake_speed *= n
    direction = randint(0, 4)

    return {
        0: (position[0], position[1] + shake_speed),    # top
        1: (position[0] - shake_speed, position[1]),    # left
        2: (position[0], position[1] - shake_speed),    # right
        3: (position[0] + shake_speed, position[1])     # bottom
    }[direction]

def shake_increasing(
    position: tuple[int, int],
    shake_speed: int,
    n: float,
    rate_function: RateFunctionArgument = RateFunctionArgument.default()
):
    """
    Creates a shake effect on the given 'position',
    by moving the element in a random position, 
    that is being increased during the animation.

    The 'shake_speed' is the amount of pixels to
    move in each movement. The value 4 is a good
    value.

    The 'n' parameter must be a normalized value
    between 0 and 1 representing the amount of
    animation that has been processed until the
    moment in which this method is executed.
    """
    shake_speed *= rate_function.get_n_value(n)
    
    return shake(position, shake_speed)

def shake_decreasing(
    position: tuple[int, int],
    shake_speed: int,
    n: float,
    rate_function: RateFunctionArgument = RateFunctionArgument.default()
):
    """
    Creates a shake effect on the given 'position',
    by moving the element in a random position, 
    that is being decreased during the animation.

    The 'shake_speed' is the amount of pixels to
    move in each movement. The value 4 is a good
    value.

    The 'n' parameter must be a normalized value
    between 0 and 1 representing the amount of
    animation that has been processed until the
    moment in which this method is executed.
    """
    shake_speed *= 1 - rate_function.get_n_value(n)
    
    return shake(position, shake_speed)

def circles(
    position: tuple[int, int],
    time_per_circle: float,
    radius: int,
    n: float,
    rate_function: RateFunctionArgument = RateFunctionArgument.default()
):
    """
    Returns the (x, y) position tuple for the moviepy '.with_position()' effect,
    for each 't' provided, that will make the element move in circles with the
    provided 'radius'. The 'radius' parameter is the distance between the origin
    and the path the clip will follow. The 'cicle_time' is the time (in seconds)
    needed for a complete circle to be completed by the movement.

    If you provide the video duration as 'cicle_time', the video will make only
    one whole circle
    """
    circle_factor = rate_function.get_n_value(n) % time_per_circle

    return position[0] + radius * np.cos((circle_factor) * 2 * pi), position[1] + radius * np.sin((circle_factor) * 2 * pi)


from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.edition.video_frame_t_helper import VideoFrameTHelper
from yta_image.converter import ImageConverter
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.parameter_validator import NumberValidator, PythonValidator
from moviepy.Clip import Clip
from imageio import imsave
from typing import Union

import numpy as np


class FrameExtractionType(Enum):
    """
    Enum to simplify the way we indicate the frame extraction
    operation we want to execute.
    """
    FRAME_TIME_MOMENT = 'frame_time_moment'
    """
    Use the frame time moments, which are the time moments 
    that belong to the time in which the frame is being
    displayed according to its duration and fps.
    """
    FRAME_INDEX = 'frame_index'
    """
    Use the frame indexes, which are the indexes of the
    frames, one after the other.
    """

class VideoFrameExtractor:
    """
    Class to simplify the process of extracting video frames by time
    or frame number.

    A moviepy clip is built by consecutive frames. The first frame is
    on t = 0, and the next frames are obtained by applying a (a, b]
    interval. This means that if we a have a video of 10 fps, each 
    frame will last 0.1s. So, considering the previous condition, the
    second frame will be at 0.1s, so you will be able to access it by
    providing a time between (0.000000000001, 0.1].
    """

    # TODO: Implement the 'output_filename' optional parameter to store
    # the frames if parameter is provided.

    @staticmethod
    def get_frame_by_index(
        video: Clip,
        index: int = 0,
        output_filename: Union[str, None] = None
    ):
        """
        Get the frame 't' of the provided 'video'. The frame number
        must be a valid one.
        """
        frame = VideoFrameExtractor.get_frame(video, FrameExtractionType.FRAME_INDEX, index)

        if output_filename:
            # TODO: Why type (?)
            #imsave(output_filename, frame.astype("uint8"))
            imsave(output_filename, frame)

        return frame

    @staticmethod
    def get_frames_by_index(
        video: Clip,
        indexes: list[float] = [0]
    ):
        """
        Get all the 't' frames of the provided 'video'. Those frame
        numbers must be valid.
        """
        if not PythonValidator.is_list(indexes):
            if NumberValidator.is_positive_number(indexes):
                indexes = [indexes]
            else:
                raise Exception('The provided "indexes" is not an array of frame numbers nor a single one.')
            
        return [
            VideoFrameExtractor.get_frame(video, FrameExtractionType.FRAME_INDEX, index)
            for index in indexes
        ]
    
    @staticmethod
    def get_frame_by_t(
        video: Clip,
        t: float = 0.0,
        output_filename: Union[str, None] = None
    ):
        """
        Get the frame in the given 't' time moment of the
        provided 'video'. The frame time must be a valid one,
        between 0 and the video duration.
        """
        frame = VideoFrameExtractor.get_frame(video, FrameExtractionType.FRAME_INDEX, t)

        if output_filename:
            # TODO: Why type (?)
            #imsave(output_filename, frame.astype("uint8"))
            imsave(output_filename, frame)

        return frame

    @staticmethod
    def get_frames_by_t(
        video: Clip,
        t: list[float] = [0.0]
    ):
        """
        Get all the frame corresponding to the time moments in
        provided 't' of the provided 'video'. All frame times
        must be valid, between 0 and the video duration.
        """
        if not PythonValidator.is_list(t):
            if NumberValidator.is_positive_number(t):
                t = [t]
            else:
                raise Exception('The provided "t" is not an array of frame times nor a single one.')
            
        return [
            VideoFrameExtractor.get_frame(video, FrameExtractionType.FRAME_TIME_MOMENT, t_)
            for t_ in t
        ]
    
    @staticmethod
    def get_frame(
        video: Clip,
        mode: FrameExtractionType = FrameExtractionType.FRAME_INDEX,
        t: Union[float, int] = 0
    ):
        """
        Get the frame corresponding to the provided 't' frame number
        or time moment. Feel free to use the moviepy .get_frame()
        method instead.

        This is my own method due to some problems with the original
        moviepy .get_frame() and because of its laxity.
        """
        video = VideoParser.to_moviepy(video)
        mode = FrameExtractionType.to_enum(mode)

        if not NumberValidator.is_positive_number(t):
            raise Exception('The provided "t" value is not a positive number.')

        t = {
            FrameExtractionType.FRAME_INDEX: VideoFrameTHelper.get_frame_t_from_frame_index(int(t), video.fps),
            FrameExtractionType.FRAME_TIME_MOMENT: float(t)
        }[mode]

        return video.get_frame(t)

    @staticmethod
    def get_all_frames(
        video: Clip
    ):
        """
        Returns all the frames from the provided 'video'.
        """
        video = VideoParser.to_moviepy(video)

        return [
            frame
            for frame in video.iter_frames()
        ]
    
    @staticmethod
    def get_first_frame(
        video: Clip
    ):
        """
        Obtain the first frame of the provided 'video' as a ndim=3
        numpy array containing the clip part (no mask) as not
        normalized values (between 0 and 255).
        """
        video = VideoParser.to_moviepy(video)

        return VideoFrameExtractor.get_frame_by_index(video, 0)
    
    @staticmethod
    def get_last_frame(
        video: Clip
    ):
        """
        Obtain the last frame of the provided 'video' as a ndim=3
        numpy array containing the clip part (no mask) as not
        normalized values (between 0 and 255).
        """
        video = VideoParser.to_moviepy(video)

        number_of_frames = VideoFrameTHelper.get_number_of_frames(video.duration, video.fps)

        return VideoFrameExtractor.get_frame_by_index(len(number_of_frames) - 1)
    
    # TODO: Would be perfect to have some methods to turn frames into
    # RGBA denormalized (0, 255) or normalized (0, 1) easier because
    # it is needed to work with images and other libraries. Those 
    # methods would iterate over the values and notice if they are in
    # an specific range so they need to be change or even if they are
    # invalid values (not even in [0, 255] range because they are not
    # rgb or rgba colors but math calculations).
    # This is actually being done by the VideoMaskHandler
    @staticmethod
    def get_frame_as_rgba_by_t(
        video: Clip,
        t: float,
        do_normalize: bool = False,
        output_filename: str = None
    ):
        """
        Gets the frame of the requested 't' time moment of the
        provided 'video' as a normalized RGBA numpy array that
        is built by joining the rgb frame (from main clip) and
        the alpha (from .mask clip), useful to detect transparent
        regions.
        """
        video = VideoParser.to_moviepy(video, do_include_mask = True)

        # We first normalize the clips
        main_frame = VideoFrameExtractor.get_frame_by_t(video, t) / 255  # RGB numpy array normalized 3d <= r,g,b
        mask_frame = VideoFrameExtractor.get_frame_by_t(video.mask, t)[:, :, np.newaxis]  # Alpha numpy array normalized 1d <= alpha
        # Combine RGB of frame and A from mask to RGBA numpy array (it is normalized)
        frame_rgba = np.concatenate((main_frame, mask_frame), axis = 2) # 4d <= r,g,b,alpha

        if output_filename:
            # TODO: Check extension
            ImageConverter.numpy_image_to_pil(frame_rgba).save(output_filename)
            # TODO: Write numpy as file image
            # Video mask is written as 0 or 1 (1 is transparent)
            # but main frame is written as 0 to 255, and the
            # 'numpy_image_to_pil' is expecting from 0 to 1
            # (normalized) instead of from 0 to 255 so it won't
            # work

        if not do_normalize:
            frame_rgba *= 255

        return frame_rgba
    